import tempfile

import pandas as pd

from caradoc import DataOutput, ExcelTable


def test_dataoutput():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test", title="test")
    assert do.sheets["test"][0].df.equals(df)


def test_dataoutput_exceltable():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    tb = ExcelTable(df, title="test")
    do.add_table(tb, "test")
    assert do.sheets["test"][0].df.equals(df)


def test_dataoutput_excel():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test", title="test")
    with tempfile.TemporaryDirectory() as tmp:
        do.write(tmp + "/test.xlsx")
        assert pd.read_excel(tmp + "/test.xlsx", sheet_name="test", skiprows=2).equals(df)


def test_dataoutput_duplicate():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test", title="test")
    do.add_table(df, "test", title="test")
    assert len(do.sheets["test"]) == 1

    do.add_table(df, "test", title="test_again")
    assert len(do.sheets["test"]) == 2


def test_dataoutput_duplicate_df():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test")
    do.add_table(df, "test")
    assert len(do.sheets["test"]) == 1

    do.add_table(df, "test", title="test_again")
    assert len(do.sheets["test"]) == 2


def test_dataoutput_excel_true_index():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]}, index=["A", "B", "C"])
    do.add_table(df, "test", title="test")
    with tempfile.TemporaryDirectory() as tmp:
        do.write(tmp + "/test.xlsx")
        assert pd.read_excel(tmp + "/test.xlsx", sheet_name="test", skiprows=2).equals(df.reset_index())


def test_dataoutput_excel_column_widths():
    df = pd.DataFrame({"a": [1, 2, 3]})
    et = ExcelTable(df, title="test")
    with tempfile.TemporaryDirectory() as tmp:
        with pd.ExcelWriter(tmp + "/test.xlsx") as writer:
            et.to_excel_table(writer, sheet_name="test", column_widths={"a": 10})

        with pd.ExcelWriter(tmp + "/test2.xlsx") as writer:
            et.to_excel_table(writer, sheet_name="test", do_column_widths=False)


def test_dataoutput_excel_no_title():
    df = pd.DataFrame({"a": [1, 2, 3]})
    et = ExcelTable(df)
    with tempfile.TemporaryDirectory() as tmp:
        with pd.ExcelWriter(tmp + "/test.xlsx") as writer:
            et.to_excel(writer, sheet_name="test")


def test_dataoutput_excel_summary_notes():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test", title="test", summary="summary", notes="notes")
    with tempfile.TemporaryDirectory() as tmp:
        do.write(tmp + "/test.xlsx")
        # length should be:
        # 1 row for title
        # 1 row for summary
        # 1 row gap
        # 4 rows for data
        # 1 row for notes
        # = 8
        assert len(pd.read_excel(tmp + "/test.xlsx", sheet_name="test", header=None)) == 8


def test_dataoutput_excel_summary_notes_double():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test", title="test", summary="summary", notes="notes")
    do.add_table(df, "test", title="test_again", summary="summary", notes="notes")
    with tempfile.TemporaryDirectory() as tmp:
        do.write(tmp + "/test.xlsx")
        # length should be:
        # 1 row for title
        # 1 row for summary
        # 1 row gap
        # 4 rows for data
        # 1 row for notes
        # x2 for second table
        # 1 row gap between tables
        # = 17
        assert len(pd.read_excel(tmp + "/test.xlsx", sheet_name="test", header=None)) == 17


def test_dataoutput_summary_notes():
    do = DataOutput()
    df = pd.DataFrame({"a": [1, 2, 3]})
    do.add_table(df, "test", title="test", summary="summary", notes="notes")
    assert do.sheets["test"][0].df.equals(df)


def test_exceltable_not_equal():
    df = pd.DataFrame({"a": [1, 2, 3]})
    et = ExcelTable(df, title="test")
    assert et != 1



from datetime import date

import pytest

from caradoc import FinancialYear


def test_financialyear_from_int():
    fy = FinancialYear.from_int(2020)
    assert fy == "2020-21"


def test_financialyear_from_date():
    fy = FinancialYear.from_date(date(2020, 1, 1))
    assert fy == "2019-20"

    fy = FinancialYear.from_date(date(2020, 4, 1))
    assert fy == "2020-21"


def test_financialyear_str():
    assert str(FinancialYear("2020-21")) == "2020-21"


def test_financialyear_repr():
    assert repr(FinancialYear("2020-21")) == "2020-21"


def test_financialyear_error():
    with pytest.raises(ValueError):
        FinancialYear("FLAM-FLIM")


def test_financialyear_equality():
    assert FinancialYear("2020-21") < FinancialYear("2021-22")
    assert FinancialYear("2020-21") <= FinancialYear("2021-22")
    assert FinancialYear("2020-21") <= FinancialYear("2020-21")
    assert FinancialYear("2020-21") == FinancialYear("2020-21")
    assert FinancialYear("2020-21") != FinancialYear("2021-22")
    assert FinancialYear("2020-21") >= FinancialYear("2020-21")
    assert FinancialYear("2025-26") >= FinancialYear("2020-21")
    assert FinancialYear("2025-26") > FinancialYear("2020-21")


def test_financialyear_equality_int():
    assert FinancialYear("2020-21") < 2021
    assert FinancialYear("2020-21") <= 2021
    assert FinancialYear("2020-21") <= 2020
    assert FinancialYear("2020-21") == 2020
    assert FinancialYear("2020-21") != 2021
    assert FinancialYear("2020-21") >= 2020
    assert FinancialYear("2025-26") >= 2020
    assert FinancialYear("2025-26") > 2020


def test_financialyear_addition():
    assert FinancialYear("2020-21") - FinancialYear("2019-20") == 1
    assert FinancialYear("2020-21") - 1 == FinancialYear("2019-20")
    assert FinancialYear("2020-21") + 1 == FinancialYear("2021-22")
    assert FinancialYear("2020-21") + 5 == FinancialYear("2025-26")


def test_financialyear_in_dict():
    fy = FinancialYear("2020-21")
    d = {fy: 1}
    assert d[fy] == 1


def test_financialyear_range():
    assert list(FinancialYear.range("2019-20", "2021-22")) == [
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
        FinancialYear("2021-22"),
    ]


def test_financialyear_range_backwards():
    assert list(FinancialYear.range("2021-22", "2019-20")) == [
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
        FinancialYear("2021-22"),
    ]


def test_financialyear_range_fy():
    assert list(FinancialYear.range(FinancialYear("2019-20"), FinancialYear("2021-22"))) == [
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
        FinancialYear("2021-22"),
    ]
    assert FinancialYear("2020-21") in FinancialYear.range("2019-20", "2021-22")


def test_financialyear_next_year():
    assert FinancialYear("2020-21").next_year() == FinancialYear("2021-22")


def test_financialyear_previous_year():
    assert FinancialYear("2020-21").previous_year() == FinancialYear("2019-20")


def test_financialyear_previous_n_years():
    assert FinancialYear("2020-21").previous_n_years() == [
        FinancialYear("2018-19"),
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
    ]
    assert FinancialYear("2020-21").previous_n_years(n_previous=1) == [
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
    ]
    assert FinancialYear("2020-21").previous_n_years(n_future=1) == [
        FinancialYear("2018-19"),
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
        FinancialYear("2021-22"),
    ]
    assert FinancialYear("2020-21").previous_n_years(n_previous=1, n_future=1) == [
        FinancialYear("2019-20"),
        FinancialYear("2020-21"),
        FinancialYear("2021-22"),
    ]


def test_financialyear_contains():
    d = date(2020, 4, 1)
    assert d in FinancialYear("2020-21")
    assert d not in FinancialYear("2019-20")
    assert d not in FinancialYear("2019-20")


def test_financialyear_contains_error():
    with pytest.raises(NotImplementedError):
        assert "Hello" in FinancialYear("2020-21")  # type: ignore


def test_financialyear_int():
    assert int(FinancialYear("2020-21")) == 2020
    assert int(FinancialYear("2019-20")) == 2019
    assert int(FinancialYear("2018-19")) == 2018


def test_financialyear_float():
    assert float(FinancialYear("2020-21")) == 2020.0
    assert float(FinancialYear("2019-20")) == 2019.0
    assert float(FinancialYear("2018-19")) == 2018.0



import re
from datetime import date
from typing import List, Union

FY_REGEX = re.compile(r"^(\d{4}) ?[\-\/] ?(\d{2,4})$")
DEFAULT_END_MONTH = 3


class FinancialYear:
    """Represents a financial year."""

    def __init__(self, fy: str, separator: str = "-") -> None:
        match = FY_REGEX.match(fy)
        if not match:
            err = "Financial year must be in the format YYYY-YY"
            raise ValueError(err)
        self.separator: str = separator

        self.year = int(match.group(1))
        self.end_year: int = self.year + 1

        self.fy = f"{self.year}{self.separator}{str(self.end_year)[-2:]}"

        self.start_date = date(self.year, 4, 1)
        self.end_date = date(self.end_year, 3, 31)

    @classmethod
    def from_int(cls, year_int: int, separator: str = "-") -> "FinancialYear":
        return cls(f"{year_int}-{str(year_int+1)[-2:]}", separator=separator)

    @classmethod
    def from_date(cls, d: date, separator: str = "-") -> "FinancialYear":
        if d.month <= DEFAULT_END_MONTH:
            return cls.from_int(d.year - 1, separator=separator)
        return cls.from_int(d.year, separator=separator)

    def __str__(self) -> str:
        return self.fy

    def __repr__(self) -> str:
        return self.fy

    def __hash__(self) -> int:
        return hash(self.fy)

    def __lt__(self, other: Union["FinancialYear", int]) -> bool:
        if isinstance(other, int):
            return self.year < other
        return self.year < other.year

    def __le__(self, other: Union["FinancialYear", int]) -> bool:
        if isinstance(other, int):
            return self.year <= other
        return self.year <= other.year

    def __eq__(self, other) -> bool:
        if isinstance(other, int):
            return self.year == other
        return str(self) == str(other)

    def __ne__(self, other) -> bool:
        if isinstance(other, int):
            return self.year != other
        return str(self) != str(other)

    def __gt__(self, other: Union["FinancialYear", int]) -> bool:
        if isinstance(other, int):
            return self.year > other
        return self.year > other.year

    def __ge__(self, other: Union["FinancialYear", int]) -> bool:
        if isinstance(other, int):
            return self.year >= other
        return self.year >= other.year

    def __sub__(self, other: Union["FinancialYear", int]) -> Union["FinancialYear", int]:
        if isinstance(other, int):
            return FinancialYear.from_int(self.year - other, separator=self.separator)
        return self.year - other.year

    def __add__(self, other: int) -> "FinancialYear":
        return FinancialYear.from_int(self.year + other, separator=self.separator)

    def __contains__(self, d: date) -> bool:
        if not isinstance(d, date):
            err = "FinancialYear can only contain dates"
            raise NotImplementedError(err)
        return self.start_date <= d <= self.end_date

    def __int__(self) -> int:
        return self.year

    def __float__(self) -> float:
        return float(self.year)

    def next_year(self) -> "FinancialYear":
        return FinancialYear.from_int(self.year + 1, separator=self.separator)

    def previous_year(self) -> "FinancialYear":
        return FinancialYear.from_int(self.year - 1, separator=self.separator)

    def previous_n_years(self, n_previous: int = 2, n_future: int = 0) -> List["FinancialYear"]:
        """Returns a list of previous and future financial years."""
        return (
            [FinancialYear.from_int(y, separator=self.separator) for y in range(self.year - n_previous, self.year)]
            + [self]
            + [
                FinancialYear.from_int(y, separator=self.separator)
                for y in range(self.year + 1, self.year + n_future + 1)
            ]
        )

    @staticmethod
    def range(  # noqa: A003
        fy: Union["FinancialYear", str], other: Union["FinancialYear", str]
    ) -> List["FinancialYear"]:
        """Returns a list of financial years between two financial years."""
        if isinstance(fy, str):
            fy = FinancialYear(fy)
        if isinstance(other, str):
            other = FinancialYear(other)
        if fy.year > other.year:
            return [FinancialYear.from_int(y, separator=fy.separator) for y in range(other.year, fy.year + 1)]
        return [FinancialYear.from_int(y, separator=fy.separator) for y in range(fy.year, other.year + 1)]



from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional, Union

import pandas as pd


class ExcelTable:
    """Represents a table in an Excel workbook.

    The table itself is a pandas DataFrame. The DataFrame
    index is not written to the Excel file.

    Allows for specifying a title, summary and notes for the table.

    # Parameters

    - `df`: pandas DataFrame
    - `title`: Optional title for the table
    - `summary`: Optional summary for the table
    - `notes`: Optional notes for the table

    # Methods

    - `to_excel_table()`: Writes just the datatable (`df`) to an Excel file
        as a Table (with filters)
    - `to_excel()`: Writes the table to an Excel file as a Table, with the
        title and summary as a header and the notes as a footer."""

    def __init__(
        self,
        df: pd.DataFrame,
        title: Optional[str] = None,
        summary: Optional[str] = None,
        notes: Optional[str] = None,
    ):
        self.df: pd.DataFrame = df
        self.title: Optional[str] = title
        self.summary: Optional[str] = summary
        self.notes: Optional[str] = notes

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExcelTable):
            return False
        if getattr(self, "title", None) or getattr(other, "title", None):
            return getattr(self, "title", None) == getattr(other, "title", None)
        return self.df.equals(other.df)

    def to_excel_table(
        self,
        writer: pd.ExcelWriter,
        sheet_name: str,
        column_widths: Optional[Dict[str, int]] = None,
        max_col_width: int = 50,
        startrow: int = 0,
        do_column_widths: bool = True,  # noqa: FBT002, FBT001
        **kwargs,
    ):
        self.df.to_excel(writer, sheet_name=sheet_name, startrow=startrow + 1, header=False, index=False, **kwargs)

        if column_widths is None:
            column_widths = {
                k: v
                for k, v in self.df.apply(lambda x: x.astype(str).str.len().max(), axis=0)
                .apply(lambda x: min(x, max_col_width))
                .to_dict()
                .items()
                if isinstance(k, str) and isinstance(v, int)
            }

        # Get the xlsxwriter workbook and worksheet objects.
        _ = writer.book
        worksheet = writer.sheets[sheet_name]

        # Get the dimensions of the dataframe.
        (max_row, max_col) = self.df.shape
        max_row = startrow + max_row

        # Create a list of column headers, to use in add_table().
        column_settings = [{"header": column} for column in self.df.columns]

        # Add the Excel table structure. Pandas will add the data.
        worksheet.add_table(startrow, 0, max_row, max_col - 1, {"columns": column_settings})

        # Make the columns wider for clarity.
        if do_column_widths:
            for col_index, column in enumerate(self.df.columns):
                column_name = str(column)
                worksheet.set_column(
                    col_index,
                    col_index,
                    min(
                        max(column_widths.get(column_name, 12), len(column_name)),
                        max_col_width,
                    ),
                )

    def to_excel(self, writer: pd.ExcelWriter, sheet_name: str, current_row: int = 0, **kwargs):
        additional_rows = 0
        to_write: list[tuple[int, int, str, Any] | tuple[int, int, str]] = []

        # write the title
        if self.title is not None:
            title_format = writer.book.add_format(  # type: ignore
                {
                    "bold": True,
                    "font_size": 15,
                }
            )
            to_write.append((current_row, 0, self.title, title_format))
            current_row += 1
            additional_rows = 1

        # write the summary
        if self.summary is not None:
            to_write.append((current_row, 0, self.summary))
            current_row += 1
            additional_rows = 1

        current_row += additional_rows

        self.to_excel_table(writer, sheet_name=sheet_name, startrow=current_row, **kwargs)
        current_row += len(self.df) + 1

        # write any notes
        if self.notes is not None:
            to_write.append((current_row, 0, self.notes))
            current_row += 1

        for args in to_write:
            writer.sheets[sheet_name].write(*args)

        return current_row + 1


class DataOutput:
    """Represents a collection of ExcelTables to be written to an Excel file.

    # Methods

    - `add_table()`: Adds a table to the DataOutput
    - `write()`: Writes the DataOutput to an Excel file"""

    def __init__(self) -> None:
        self.sheets: DefaultDict[str, List[ExcelTable]] = defaultdict(lambda: [])

    def add_table(self, df: Union[pd.DataFrame, ExcelTable], sheet: str, **kwargs) -> pd.DataFrame:
        """Adds a table to the DataOutput.

        Note that adding a dataframe with the same title as an existing table will
        overwrite the existing table.

        # Parameters

        - `df`: pandas DataFrame or ExcelTable
        - `sheet`: Name of the sheet to write the table to
        - `kwargs`: Additional arguments to pass to ExcelTable constructor.
            Only used if `df` is a DataFrame. The accepted arguments are `title`,
            `summary` and `notes` - all of which should be strings.
        """
        if isinstance(df, pd.DataFrame):
            if isinstance(df.index, pd.RangeIndex):
                et = ExcelTable(df, **kwargs)
            else:
                et = ExcelTable(df.reset_index(), **kwargs)
        else:
            et = df
        self.sheets[sheet] = [t for t in self.sheets[sheet] if t != et] + [et]
        return et.df

    def write(self, file_name: str) -> None:
        """Writes the DataOutput to an Excel file.

        # Parameters

        - `file_name`: Name of the file to write to"""
        with pd.ExcelWriter(file_name, engine="auto") as writer:
            for sheet_name, tables in self.sheets.items():
                current_row = 0
                for table in tables:
                    current_row = table.to_excel(writer, sheet_name=sheet_name, current_row=current_row)
